/* 
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 539 $ $LastChangedDate: 2002-08-25 13:25:13 +0200 (Sun, 25 Aug 2002) $
 *
 * Addons library for Gont, Xml module implementation.
 */

open Xml

struct doc {
	node root;
}

type c_doc;
type c_node;

typedef String::cstring utf_string;

extern_c c_doc xmlParseFile(utf_string name);
extern_c c_doc xmlParseDoc(utf_string xml);

type c_node;
type c_ns;
type c_attr;

extern_c c_doc xmlNewDoc(utf_string version);
extern_c c_node xmlNewDocNode(c_doc, c_ns, utf_string name, utf_string value);
extern_c c_node xmlNewChild(c_node parent, c_ns, utf_string name, 
			    utf_string value);
extern_c c_node xmlNewText(utf_string value);
extern_c c_node xmlAddChild(c_node parent, c_node n);
extern_c c_attr xmlSetProp(c_node n, utf_string name, utf_string value);
extern_c void xmlFreeDoc(c_doc);

extern_c string Xml___xmlDocDumpMemory(c_doc, bool);
extern_c node Xml___convert_document(c_doc, bool);
extern_c void Xml___set_root(c_doc, c_node);

utf_string null_string;
c_ns null_ns;

// we cheat a bit here :<
utf_string to_utf(string s)
{
	return Hacks::cast(String::to_cstring(s));
}

doc make_doc(node n)
{
	return ({root = n});
}

doc convert_document(c_doc cd, bool skip_spaces)
{
	node root;

	try {
		root = Xml___convert_document(cd, skip_spaces);
	} with {
	case Invalid_argument[_]:
		raise Parse_error["cannot parse document"];
	} finally {
		xmlFreeDoc(cd);
	}

	return make_doc(root);
}

doc parse_file(string filename, bool ss)
{
	return convert_document(xmlParseFile(to_utf(filename)), ss);
}

doc parse_string(string xml, bool ss)
{
	return convert_document(xmlParseDoc(to_utf(xml)), ss);
}

node root_node(doc d)
{
	return d.root;
}

void set_attrs(c_node n, <*[string, string]>List::t attrs)
{
	void set_one(*[string, string] pair)
	{
		let [name, value] = pair in
		xmlSetProp(n, to_utf(name), to_utf(value));
	}

	List::iter(set_one, attrs);
}

void process_node(c_node parent, node n)
{
	c_node self;
	
	switch n {
	case Element[name, children, attrs]:
		self = xmlNewChild(parent, null_ns, to_utf(name), null_string);
		set_attrs(self, attrs);
		List::iter((fun void (node x) (process_node(self, x))), children);
	
	case Text[s]:
		self = xmlNewText(to_utf(s));
		xmlAddChild(parent, self);
	}
}

c_doc doc_to_c_doc(doc d)
{
	c_doc ret;
	c_node root;
	
	switch root_node(d) {
	case Element[name, children, attrs]:
		ret = xmlNewDoc(to_utf("1.0"));
		root = xmlNewDocNode(ret, null_ns, to_utf(name), null_string);
		Xml___set_root(ret, root);
		set_attrs(root, attrs);
		List::iter((fun void (node n) (process_node(root, n))), children);
	case _:
		raise Invalid_argument["Xml::doc_to_c_doc"];
	}
	
	
	return ret;
}

string string_of_doc(doc d)
{
	c_doc cd = doc_to_c_doc(d);
	string r = Xml___xmlDocDumpMemory(cd, false);
	xmlFreeDoc(cd);
	return r;
}

string string_of_doc_indented(doc d)
{
	c_doc cd = doc_to_c_doc(d);
	string r = Xml___xmlDocDumpMemory(cd, true);
	xmlFreeDoc(cd);
	return r;
}

// TODO: there might be direct implementation in future...
void save_to_stream(doc d, Stdio::FILE f)
{
	string s = string_of_doc(d);
	Stdio::fputs(f, s);
}

void save_to_stream_indented(doc d, Stdio::FILE f)
{
	string s = string_of_doc_indented(d);
	Stdio::fputs(f, s);
}

void save_to_file(doc d, string name)
{
	Stdio::FILE f = Stdio::fopen(name, "w");
	try {
		save_to_stream(d, f);
	} finally {
		Stdio::fclose(f);
	}
}

void save_to_file_indented(doc d, string name)
{
	Stdio::FILE f = Stdio::fopen(name, "w");
	try {
		save_to_stream_indented(d, f);
	} finally {
		Stdio::fclose(f);
	}
}

node make_text_node(string s)
{
	return Text[s];
}

node make_element_node(string name, <node>List::t children, 
		       <*[string, string]>List::t attrs)
{
	return Element[name, children, attrs];
}

node find_element(string name, node n)
{
	switch n {
	case Xml::Element[_, children, _]:
		return List::find((fun bool (node x) {
					switch x {
					case Element[s, _, _]:
					 	return s == name;
					case _:
						return false;
					}
				  }), children);
	case _:
		raise Not_found;
	}
}

string prop_value(string name, node n)
{
	switch n {
	case Xml::Element[_, _, props]:
		let [_, v] = List::find(
			(fun bool (*[string, string] p)
				{let [s, _] = p in return s == name;}),
			props) in
		return v;
	case _:
		raise Not_found;
	}
			        
}

string prop_value_default(string name, node n, string def)
{
	try {
		return prop_value(name, n);
	} with {
	case Not_found:
		return def;
	}
}
